#ifndef __CORE_SW2_NODE_H_
#define __CORE_SW2_NODE_H_

#define KERNEL_OFFSET	0xfffffc0000000000UL


#define SWVMM_SHMEM_START_ADDR	0x600000		/* 6M */
#define SWVMM_SHMEM_SIZE	0x100000 

#define CG_NUM		4	
#define CG_ONLINE_IOADDR	((0x1UL<<39) | (0x5UL<<36) | (0x0UL<<7) | KERNEL_OFFSET)     

#define CORE_SLEEP_STAT_IOADDR	((0x1UL<<39) | (0x6UL<<36) | (0x1UL<<7) | KERNEL_OFFSET)
#define CORE_START_IOADDR	((0x1UL<<39) | (0x5UL<<36) | (0xfUL<<7) | KERNEL_OFFSET)

#define SW2_PCI_MEM_UNPRE 	0x0UL

#ifdef CONFIG_SWICH
#define SW2_PCI_MEM_PRE         0x200000000UL
#define SW2_PCI_MEM_PRI_SIZE    0x200000000UL 
#else
#define SW2_PCI_MEM_PRE         0x3e0000000UL
#define SW2_PCI_MEM_PRI_SIZE    0x20000000UL 
#endif

#define SW2_PCI_IO_PIU0 	0x400000000UL
#define SW2_PCI_IO_PIU1 	0x500000000UL
#define SW2_PCI_IO_EP 		0x600000000UL
#define SW2_PCI_EP_CONFIG 	0xa00000000UL
#define SW2_PCI_MEM_SIZE        0x20000000UL
#define SW2_PCI_IO_SIZE         0xf00000UL   
#define SW2_PCI_IO_RC 		0x800000000UL

/* Define each pcie base address */
#define SW2_LEGACY_IO		0xc600000000

#define SW2_IO_PCIE(n)			(0xc000000000UL + (0x3UL << 36) * (n))
#define SW2_PCIE_IO_BASE(n)		(SW2_IO_PCIE(n) + SW2_PCI_IO_EP)
#define SW2_PCIE_MEM_BASE(n)		(SW2_IO_PCIE(n) + (0xeUL << 28))
#define SW2_PCIE_MEM_PRI_BASE(n)	(SW2_IO_PCIE(n) + SW2_PCI_MEM_PRE)

#define MSI_FIRST_VECTOR 15

/* PIU internal register define */
#define ERR_REPORT 		0x0
#define CXCPL_DEBUG_INFO	0x5
#define MSI_ADDR		0x9

/* Define each PCIE start scanning bus number */
#define SW2_PCI0_BUS	0x1
#define SW2_PCI1_BUS	0x20


#define SW2_PCI_DEBUG 0
#if SW2_PCI_DEBUG
#define PCIINFO(fmt, args...)	printk(fmt, ##args)
#else
#define PCIINFO(fmt, args...) 
#endif

#ifndef __EXTERN_INLINE
#define __EXTERN_INLINE extern inline
#endif

static inline int is_cg_alive(int cgid)
{
       return ((((*(unsigned long *)CORE_START_IOADDR) >> (cgid * 4)) & 0xf) ? 1 : 0);
}

static inline unsigned long get_RC_config(int index)
{
	return SW2_IO_PCIE(index) + KERNEL_OFFSET + SW2_PCI_IO_RC;
}

static inline unsigned long get_EP_PCI_config(int index)
{
	return SW2_IO_PCIE(index) + KERNEL_OFFSET + SW2_PCI_EP_CONFIG + 0x100000000;
}

static inline unsigned long get_PIU_PCI_IOR0(int index)
{
	return SW2_IO_PCIE(index) + KERNEL_OFFSET + SW2_PCI_IO_PIU0;
}

static inline unsigned long get_PIU_PCI_IOR1(int index)
{
	return SW2_IO_PCIE(index) + KERNEL_OFFSET + SW2_PCI_IO_PIU1;
}

static inline unsigned long get_PCI_config(int index)
{
	return SW2_IO_PCIE(index) + KERNEL_OFFSET + SW2_PCI_EP_CONFIG;
}

extern inline void __iomem *sw2_ioportmap(unsigned long addr)
{
#if 0
	if ( addr < 0x100000)
		addr = addr + SW2_LEGACY_IO;
	return (void __iomem *)(addr + KERNEL_OFFSET);
#else
	return sw_mv.mv_ioportmap(addr); 
#endif
}

static inline void __iomem *sw2_iounmap(volatile void __iomem *addr)
{
	unsigned long tmp = (unsigned long) addr;	
	return (void __iomem *)(tmp - KERNEL_OFFSET);
}

static inline void __iomem *sw2_ioremap(unsigned long addr, unsigned long size)
{
	unsigned long tmp = addr + KERNEL_OFFSET;
	return (void __iomem *)(tmp);
}

static inline int sw2_is_ioaddr(unsigned long addr)
{
	return addr >= (KERNEL_OFFSET + SW2_IO_PCIE(0));
}

static inline int sw2_is_mmio(const volatile void __iomem *xaddr)
{
	unsigned long addr = (unsigned long) xaddr;
	return (addr & 0x100000000UL) == 0;
}


#undef __IO_PREFIX
#define __IO_PREFIX             sw2

#define sw2_trivial_rw_bw   1
#define sw2_trivial_rw_lq   1
#define sw2_trivial_io_bw   1
#define sw2_trivial_io_lq   1

#include <asm/io_trivial.h>

static inline void
sw2_read_rc_conf(int pci_index, unsigned int conf_offset, volatile unsigned int *data)
{
	unsigned long addr;

	addr = get_RC_config(pci_index) | conf_offset;
	PCIINFO("RC config read, address = %#lx,", addr); 
	*data = *(unsigned int *)addr;
	mb();
	PCIINFO("value = %#x.\n", *data); 
}


static inline void
sw2_write_rc_conf(int pci_index, unsigned int conf_offset, unsigned int data)
{
	unsigned long addr;

	addr = get_RC_config(pci_index) | conf_offset;
	PCIINFO("RC config write, address = %#lx, data = %#x.\n", addr, data); 
	*(unsigned int *)addr = data;
	mb();
}

static inline void
sw2_write_piu_ior0(int pci_index, unsigned int reg, unsigned long data)
{
	unsigned long addr; 
	
	addr = get_PIU_PCI_IOR0(pci_index) + (reg << 7);
	PCIINFO("PIU0 write, reg = %#x, address = %#lx, data = %#lx.\n", reg, addr, data); 
	*(unsigned long __iomem*)addr = data;
	mb();
}

static inline void
sw2_read_piu_ior1(int pci_index, unsigned int reg, volatile unsigned long *value)
{
	unsigned long addr; 
	
	addr = get_PIU_PCI_IOR1(pci_index) + (reg << 7);
	PCIINFO("PIU1 read, reg = %#x, address = %#lx,", reg, addr); 
	*value = *(unsigned long __iomem*)addr;
	mb();
	PCIINFO("value = %#lx.\n", *value);
}
static inline void
sw2_write_piu_ior1(int pci_index, unsigned int reg, unsigned long data)
{
	unsigned long addr; 
	
	addr = get_PIU_PCI_IOR1(pci_index) + (reg << 7);
	PCIINFO("PIU1 write, reg = %#x, address = %#lx, data = %#lx.\n", reg, addr, data); 
	*(unsigned long __iomem*)addr = data;
	mb();
}

static inline void
sw2_read_piu_ior0(int pci_index, unsigned int reg, volatile unsigned long *value)
{
	unsigned long addr; 
	
	addr = get_PIU_PCI_IOR0(pci_index) + (reg << 7);
	PCIINFO("PIU0 read, reg = %#x, address = %#lx,", reg, addr); 
	*value = *(unsigned long __iomem*)addr;
	mb();
	PCIINFO("value = %#lx.\n", *value);
}

#endif /* __CORE_SW2_NODE_H_ */
