#ifndef __SW_64_PCI_H
#define __SW_64_PCI_H

#ifdef __KERNEL__

#include <linux/spinlock.h>
#include <linux/dma-mapping.h>
#include <linux/scatterlist.h>
#include <asm/machvec.h>
#include <asm-generic/pci-bridge.h>

/*
 * The following structure is used to manage multiple PCI busses.
 */

struct pci_dev;
struct pci_bus;
struct resource;
struct pci_iommu_arena;
struct sw_iommu;
struct page;

/* A controller.  Used to manage multiple PCI busses.  */

struct pci_controller {
	struct pci_controller *next;
        struct pci_bus *bus;
	struct resource *io_space;
	struct resource *mem_space;
	struct resource *pre_mem_space;
	/* The following are for reporting to userland.  The invariant is
	   that if we report a BWX-capable dense memory, we do not report
	   a sparse memory at all, even if it exists.  */
	unsigned long sparse_mem_base;
	unsigned long dense_mem_base;
	unsigned long sparse_io_base;
	unsigned long dense_io_base;

	/* This one's for the kernel only.  It's in KSEG somewhere.  */
	unsigned long config_space_base;

	unsigned int index;
	/* For compatibility with current (as of July 2003) pciutils
	   and XFree86. Eventually will be removed. */
	unsigned int need_domain_info;

	struct pci_iommu_arena *sg_pci;
	struct pci_iommu_arena *sg_isa;

#ifdef CONFIG_IOMMU
	struct sw_iommu *sw_sg_pci; /* For SW4A IOMMU */
#endif

	void *sysdata;
};

/* Override the logic in pci_scan_bus for skipping already-configured
   bus numbers.  */

#define pcibios_assign_all_busses()	1

#define PCIBIOS_MIN_IO		sw_mv.min_io_address
#define PCIBIOS_MIN_MEM		sw_mv.min_mem_address

extern void pcibios_set_master(struct pci_dev *dev);

/* IOMMU controls.  */

/* The PCI address space does not equal the physical memory address space.
   The networking and block device layers use this boolean for bounce buffer
   decisions.  */
#define PCI_DMA_BUS_IS_PHYS  0

#ifdef CONFIG_PCI

/* implement the pci_ DMA API in terms of the generic device dma_ one */
#include <asm-generic/pci-dma-compat.h>

#endif

/* TODO: integrate with include/asm-generic/pci.h ? */
static inline int pci_get_legacy_ide_irq(struct pci_dev *dev, int channel)
{
	return channel ? 15 : 14;
}

#define pci_domain_nr(bus) ((struct pci_controller *)(bus)->sysdata)->index

static inline int pci_proc_domain(struct pci_bus *bus)
{
	struct pci_controller *hose = bus->sysdata;
	return hose->need_domain_info;
}

#endif /* __KERNEL__ */

/* Values for the `which' argument to sys_pciconfig_iobase.  */
#define IOBASE_HOSE		0
#define IOBASE_SPARSE_MEM	1
#define IOBASE_DENSE_MEM	2
#define IOBASE_SPARSE_IO	3
#define IOBASE_DENSE_IO		4
#define IOBASE_ROOT_BUS		5
#define IOBASE_FROM_HOSE	0x10000

extern struct pci_dev *isa_bridge;

extern int pci_legacy_read(struct pci_bus *bus, loff_t port, u32 *val,
			   size_t count);
extern int pci_legacy_write(struct pci_bus *bus, loff_t port, u32 val,
			    size_t count);
extern int pci_mmap_legacy_page_range(struct pci_bus *bus,
				      struct vm_area_struct *vma,
				      enum pci_mmap_state mmap_state);
extern void pci_adjust_legacy_attr(struct pci_bus *bus,
				   enum pci_mmap_state mmap_type);
#define HAVE_PCI_LEGACY	1

extern int pci_create_resource_files(struct pci_dev *dev);
extern void pci_remove_resource_files(struct pci_dev *dev);

/* To support multiple MSI override arch_setup_msi_irqs() */
#ifndef CONFIG_VSTAROS_GUEST
#define arch_setup_msi_irqs sw_setup_msi_irqs
#else
#define arch_setup_msi_irqs vstaros_setup_msi_irqs
#endif


#define PCI_VENDOR_ID_JN          0x5656
#define PCI_DEVICE_ID_JN_PCIESW   0x1000
#define PCI_DEVICE_ID_JN_PCIEUSIP 0x1200
#define PCI_DEVICE_ID_JN_PCIE2PCI 0x1314

#define PCI_DEVICE_ID_JN_USBEHCI  0x0052
#define PCI_DEVICE_ID_JN_USBOHCI0 0x0050
#define PCI_DEVICE_ID_JN_USBOHCI1 0x0051
#define PCI_DEVICE_ID_JN_AC97     0x0060
#define PCI_DEVICE_ID_JN_VGA      0x0100
#define PCI_DEVICE_ID_JN_SATA     0x0110
#define PCI_DEVICE_ID_JN_GMAC0    0x0120
#define PCI_DEVICE_ID_JN_GMAC1    0x0121
#define PCI_DEVICE_ID_JN_TCM      0x0160
#define PCI_DEVICE_ID_JN_I2C      0x0240
#define PCI_DEVICE_ID_JN_SPI      0x0250
#define PCI_DEVICE_ID_JN_XBUS     0x0260
#define PCI_DEVICE_ID_JN_PMDEBUG  0x0270
#define PCI_DEVICE_ID_JN_BT       0x0280
#define PCI_DEVICE_ID_JN_KCS      0x0290
#define PCI_DEVICE_ID_JN_UART0    0x0300
#define PCI_DEVICE_ID_JN_UART1    0x0301
#define PCI_DEVICE_ID_JN_PS20     0x0310
#define PCI_DEVICE_ID_JN_PS21     0x0311



#define NR_IRQ_VECTORS NR_IRQS

#define LAST_DEVICE_VECTOR  31

#define PCITODMA_OFFSET	0x0    /*0 offset*/

#endif /* __SW_64_PCI_H */
