#ifndef _SW_SPINLOCK_H
#define _SW_SPINLOCK_H

#ifdef  CONFIG_NUMA

#include <asm/spinlock_types.h>
#include <asm/barrier.h>
#include <asm/processor.h>
#include <linux/irqflags.h>
static inline int ldw_set_return(unsigned long * addr)
{
        int tmp;
        __asm__ __volatile__(
        "1:     ldi %0,%1\n"
        "       ldw_set %0,0(%0)\n"        //取出锁地址内容，并对锁地址置1
		"		memb\n"
        : "=&r" (tmp), "=m" (*addr)
        : "m"(*addr) : "memory");
        return tmp;
}       

#define arch_spin_lock_flags(lock, flags) arch_spin_lock(lock)
#define arch_spin_is_locked(x)	((x)->lock != 0)
#define arch_spin_unlock_wait(x) \
		do { cpu_relax(); } while ((x)->lock)

static inline void arch_spin_unlock(arch_spinlock_t * lock)
{
	mb();
	lock->lock = 0;
}

static inline void arch_spin_lock(arch_spinlock_t * lock)
{

        long tmp;
        __asm__ __volatile__(
        "1:     ldi %0,%1\n"
        "       ldw_set %0,0(%0)\n"     //取出锁地址内容，并对锁地址置1
        "       bne     %0,1b\n"        //如果不返回0（为1），则表示锁已被占住  
		"		memb\n"
        : "=&r" (tmp), "=m" (lock->lock)
        : "m"(lock->lock) : "memory");
}

static inline int arch_spin_trylock(arch_spinlock_t *lock)
{
        int ret;
        ret = ldw_set_return((unsigned long *)&(lock->lock)) != 1;
        return ret;
}

/***********************************************************/



/*
 * Read-write spinlocks, allowing multiple readers but only one writer.
 * Linux rwlocks are unfair to writers; they can be starved for an indefinite
 * time by readers.  With care, they can also be taken in interrupt context.
 *
 * In the PA-RISC implementation, we have a spinlock and a counter.
 * Readers use the lock to serialise their access to the counter (which
 * records how many readers currently hold the lock).
 * Writers hold the spinlock, preventing any readers or other writers from
 * grabbing the rwlock.
 */

/* Note that we have to ensure interrupts are disabled in case we're
 * interrupted by some other code that wants to grab the same read lock */
static  __inline__ void arch_read_lock(arch_rwlock_t *rw)
{
	unsigned long flags;
	local_irq_save(flags);
	arch_spin_lock_flags(&rw->lock, flags);
	rw->counter++;
	arch_spin_unlock(&rw->lock);
	local_irq_restore(flags);
}

/* Note that we have to ensure interrupts are disabled in case we're
 * interrupted by some other code that wants to grab the same read lock */
static  __inline__ void arch_read_unlock(arch_rwlock_t *rw)
{
	unsigned long flags;
	local_irq_save(flags);
	arch_spin_lock_flags(&rw->lock, flags);
	rw->counter--;
	arch_spin_unlock(&rw->lock);
	local_irq_restore(flags);
}

/* Note that we have to ensure interrupts are disabled in case we're
 * interrupted by some other code that wants to grab the same read lock */
static __inline__ int arch_read_trylock(arch_rwlock_t *rw)
{
	unsigned long flags;
 retry:
	local_irq_save(flags);
	if (arch_spin_trylock(&rw->lock)) {
		rw->counter++;
		arch_spin_unlock(&rw->lock);
		local_irq_restore(flags);
		return 1;
	}

	local_irq_restore(flags);
	/* If write-locked, we fail to acquire the lock */
	if (rw->counter < 0)
		return 0;

	/* Wait until we have a realistic chance at the lock */
	while (arch_spin_is_locked(&rw->lock) && rw->counter >= 0)
		cpu_relax();

	goto retry;
}

/* Note that we have to ensure interrupts are disabled in case we're
 * interrupted by some other code that wants to read_trylock() this lock */
static __inline__ void arch_write_lock(arch_rwlock_t *rw)
{
	unsigned long flags;
retry:
	local_irq_save(flags);
	arch_spin_lock_flags(&rw->lock, flags);

	if (rw->counter != 0) {
		arch_spin_unlock(&rw->lock);
		local_irq_restore(flags);

		while (rw->counter != 0)
			cpu_relax();

		goto retry;
	}

	rw->counter = -1; /* mark as write-locked */
	mb();
	local_irq_restore(flags);
}

static __inline__ void arch_write_unlock(arch_rwlock_t *rw)
{
	rw->counter = 0;
	arch_spin_unlock(&rw->lock);
}

/* Note that we have to ensure interrupts are disabled in case we're
 * interrupted by some other code that wants to read_trylock() this lock */
static __inline__ int arch_write_trylock(arch_rwlock_t *rw)
{
	unsigned long flags;
	int result = 0;

	local_irq_save(flags);
	if (arch_spin_trylock(&rw->lock)) {
		if (rw->counter == 0) {
			rw->counter = -1;
			result = 1;
		} else {
			/* Read-locked.  Oh well. */
			arch_spin_unlock(&rw->lock);
		}
	}
	local_irq_restore(flags);

	return result;
}

/*
 * read_can_lock - would read_trylock() succeed?
 * @lock: the rwlock in question.
 */
static __inline__ int arch_read_can_lock(arch_rwlock_t *rw)
{
	return rw->counter >= 0;
}

/*
 * write_can_lock - would write_trylock() succeed?
 * @lock: the rwlock in question.
 */
static __inline__ int arch_write_can_lock(arch_rwlock_t *rw)
{
	return !rw->counter;
}

#define arch_read_lock_flags(lock, flags) arch_read_lock(lock)
#define arch_write_lock_flags(lock, flags) arch_write_lock(lock)

#else      /* -----  not CONFIG_NUMA  ----- */

#include <linux/kernel.h>
#include <asm/current.h>

/*
 * Simple spin lock operations.  There are two variants, one clears IRQ's
 * on the local processor, one does not.
 *
 * We make no fairness assumptions. They have a cost.
 */

#define arch_spin_lock_flags(lock, flags) arch_spin_lock(lock)
#define arch_spin_is_locked(x)	((x)->lock != 0)
#define arch_spin_unlock_wait(x) \
		do { cpu_relax(); } while ((x)->lock)

static inline int arch_spin_value_unlocked(arch_spinlock_t lock)
{
	return lock.lock == 0;
}

static inline void arch_spin_unlock(arch_spinlock_t * lock)
{
	mb();
	lock->lock = 0;
}

static inline void arch_spin_lock(arch_spinlock_t * lock)
{
	long tmp, addr;

	__asm__ __volatile__(
        "       ldi     %2, %1\n" 
        "1:     lldw    %0, 0(%2)\n"    //锁读：获取lock->lock的值，lock_valid=1
        "       xor     %0, 1, %0\n"    //判断lock->lock的值,准备lock_flag的值
        "       wr_f    %0\n"           //写lock_flag标记：如果lock->lock==0，则置lock_flag=1;否则lock_flag=0;
        "       ldi     %0, 1\n"        //准备往lock->lock里写1
        "       lstw    %0, 0(%2)\n"    //条件写：若LOCK_MATCH满足，则往lock->lock写1
        "       rd_f    %0\n"           //读lock_success标记：如果条件写成功，则%0=1，否则%0=0
        "       beq     %0, 2f\n"       //判断是否上锁成功
        "       memb\n"
        ".subsection 2\n"
        "2:     ldw     %0, 0(%2)\n"    //读操作系统锁自旋锁标志
        "       bne     %0, 2b\n"       //如果不空闲，则等待空闲
        "       br      1b\n"
        ".previous"
        : "=&r" (tmp), "=m" (lock->lock), "=&r" (addr)
	: "m"(lock->lock) : "memory");
}

static inline int arch_spin_trylock(arch_spinlock_t *lock)
{
	return !test_and_set_bit(0, &lock->lock);
}

/***********************************************************/

static inline int arch_read_can_lock(arch_rwlock_t *lock)
{
	return (lock->lock & 1) == 0;
}

static inline int arch_write_can_lock(arch_rwlock_t *lock)
{
	return lock->lock == 0;
}

static inline void arch_read_lock(arch_rwlock_t *lock)
{
	long regx, regy, addr;

	__asm__ __volatile__(
        "       ldi     %3, %0\n"
        "1:     lldw    %1, 0(%3)\n"    //锁读
        "       xor     %1, 1, %2\n"    //如果最低位为0，则允许条件写；否则不允许
        "       wr_f    %2\n"           //写lock_flag标志
        "       subw    %1, 2, %1\n"    //高位读锁数加1
        "       lstw    %1, 0(%3)\n"    //条件写
        "       rd_f    %2\n"           //读lock_success标志
        "       beq     %2,6f\n"        //若条件写失败则重试
        "       memb\n"
        ".subsection 2\n"
        "6:     ldw     %1, 0(%3)\n"
        "       blbs    %1, 6b\n"
        "       br      1b\n"
        ".previous"
        : "=m" (*lock), "=&r" (regx), "=&r" (regy), "=&r" (addr)	
	: "m" (*lock) : "memory");
}

static inline void arch_write_lock(arch_rwlock_t *lock)
{
	long regx, addr;

	__asm__ __volatile__(
	"       ldi     %2, %0\n"
        "1:     lldw    %1, 0(%2)\n"            //锁读lock值
        "       seleq   %1, 1, $31, %1\n"       //若为lock=0则置lock_flag=1,否则lock_flag=0
        "       wr_f    %1\n"                   //写lock_flag标志
        "       ldi     %1, 1\n"
        "       lstw    %1, 0(%2)\n"            //条件写：值lock=1上写锁
        "       rd_f    %1\n"
        "       beq     %1, 6f\n"               //判断是否上锁成功
        "       memb\n"
        ".subsection 2\n"
        "6:     ldw     %1, 0(%2)\n"
        "       bne     %1, 6b\n"
        "       br      1b\n"
        ".previous"
        : "=m" (*lock), "=&r" (regx), "=&r" (addr)
	: "m" (*lock) : "memory");
}

static inline int arch_read_trylock(arch_rwlock_t * lock)
{
	long regx, addr;
	int success;

	__asm__ __volatile__(
        "       ldi     %3, %0\n"
        "1:     lldw    %1, 0(%3)\n"            //锁读lock值
        "       sellbc  %1, 1, $31, %2\n"       //判断是否write_locked: 如果是则success=lock_flag=0;否则success=lock_flag=1
        "       wr_f    %2\n"                   //写lock_flag
        "       subw    %1, 2, %1\n"            //读锁数加1
        "       lstw    %1, 0(%3)\n"            //条件写
        "       rd_f    %1\n"                   //读lock_success标志
        "       bne     %1, 2f\n"               //如果lock_success=1，则返回
        "       bne     %2, 6f\n"               //如果success=1，则重新尝试一遍
        "2:     memb\n"
        ".subsection 2\n"
        "6:     br      1b\n"
        ".previous"
        : "=m" (*lock), "=&r" (regx), "=&r" (success), "=&r" (addr)
	: "m" (*lock) : "memory");

	return success;
}

static inline int arch_write_trylock(arch_rwlock_t * lock)
{
	long regx, addr;
	int success;

	__asm__ __volatile__(
        "       ldi     %3, %0\n"
        "1:     lldw    %1, 0(%3)\n"            //锁读lock值
        "       seleq   %1, 1, $31, %2\n"       //判断是否有读锁和写锁: 如果无则success=lock_flag=1;否则success=lock_flag=0
        "       wr_f    %2\n"                   //写lock_flag标志
        "       ldi     %1, 1\n"                
        "       lstw    %1, 0(%3)\n"            //条件写lock为1
        "       rd_f    %1\n"                   //读lock_success标志
        "       bne     %1, 2f\n"               //如果lock_success成功，则返回
        "       bne     %2, 6f\n"               //如果success=1，则重新尝试一遍
        "2:     memb\n"
        ".subsection 2\n"
        "6:     br      1b\n"
        ".previous"
        : "=m" (*lock), "=&r" (regx), "=&r" (success), "=&r" (addr)
	: "m" (*lock) : "memory");

	return success;
}

static inline void arch_read_unlock(arch_rwlock_t * lock)
{
	long regx, regy, addr;
	
	__asm__ __volatile__(
        "       memb\n"
        "       ldi     %3, %0\n"
        "1:     lldw    %1, 0(%3)\n"    //锁读lock值
        "       ldi     %2, 1\n"        
        "       wr_f    %2\n"           //写lock_flag=1
        "       addw    %1, 2, %1\n"    //读锁共享数减1
        "       lstw    %1, 0(%3)\n"    //条件写
        "       rd_f    %1\n"           //读lock_success标志
        "       beq     %1, 6f\n"       //若lock_success=0，则重新unlock
        ".subsection 2\n"
        "6:     br      1b\n"
        ".previous"
        : "=m" (*lock), "=&r" (regx), "=&r" (regy), "=&r" (addr)
	: "m" (*lock) : "memory");
}

static inline void arch_write_unlock(arch_rwlock_t * lock)
{
	mb();
	lock->lock = 0;
}

#define arch_read_lock_flags(lock, flags) arch_read_lock(lock)
#define arch_write_lock_flags(lock, flags) arch_write_lock(lock)

#endif     /* -----  not CONFIG_NUMA  ----- */
#endif /* _SW_SPINLOCK_H */
