/*
 * =====================================================================================
 *       Filename:  evtchn.h
 *    Description:  
 *        Created:  2009年02月10日 20时28分11秒
 * =====================================================================================
 */


#ifndef  _SWVMM_EVTCHN_H_
#define  _SWVMM_EVTCHN_H_

#include <linux/netdevice.h>
#include <linux/kernel.h>
#include <linux/linkage.h>
#include <linux/preempt.h>
#include <linux/cpumask.h>
#include <linux/sched.h>
#include <asm/atomic.h>
#include <asm/errno.h>

typedef int evtreturn_t;

#define EVT_NONE (0)
#define EVT_HANDLED (1)
#define EVT_RETVAL(x) ((x) != 0)

#define ACTUAL_NR_EVTS 64
#define NR_EVTS ACTUAL_NR_EVTS

struct evtaction {
	evtreturn_t (*handler)(int, void *, struct pt_regs *);
	unsigned long flags;
	const char *name;
	void *evt_id;
	struct evtaction *next;
	cpumask_t mask;
	int evt_index;
};


typedef struct evt_desc {
	struct evtaction *action;	/* EVT action list */
	unsigned int status;		/* EVT status */
	spinlock_t lock;
} ____cacheline_aligned evt_desc_t;

extern evt_desc_t evt_desc [NR_EVTS];



/*
 * EVT line status.
 */
#define EVT_INPROGRESS	1	/* EVT handler active - do not enter! */
#define EVT_DISABLED	2	/* EVT disabled - do not enter! */
#define EVT_PENDING	4	/* EVT pending - replay on enable */
#define EVT_REPLAY	8	/* EVT has been replayed but not acked yet */
#define EVT_AUTODETECT	16	/* EVT is being autodetected */
#define EVT_WAITING	32	/* EVT not yet seen - for autodetection */
#define EVT_LEVEL	64	/* EVT level triggered */
#define EVT_MASKED	128	/* EVT masked - shouldn't be seen again */

#define EVT_NET_IRQ     0
#define EVT_NET1_IRQ     1
#define EVT_NET2_IRQ     2
#define EVT_NET3_IRQ    3

#define HCA_CMD_CALL    5    //For HCA command call
#define HCA_EVENT_CALL  8    //For HCA command call

#define EVT_NET_NOVALID  9 
#define EVT_NET_NOVALID1 10
#define EVT_NET_NOVALID2 11
#define EVT_NET_VALID 	12
#define EVT_NET_VALID1 	13
#define EVT_NET_VALID2	14


#define EVT_SLEEP       16
#define EVT_SLEEP1       17
#define EVT_SLEEP2       18
#define EVT_SLEEP3       19
#define EVT_MIGRATION_INIT 20
#define EVT_MIGRATION_TRANSFER_MEM 21
#define EVT_MIGRATION_SHUTDOWN_CG 22

struct netdrv_private {
	struct net_device_stats stats;
	spinlock_t lock;
	wait_queue_head_t thr_wait;

	char	valid;		// Card is useable
	spinlock_t	send_lock;
	spinlock_t	recv_lock;
	int		send_next;
	int		recv_next;
	u64	*send_bufaddr;	// SOC's Send_buf addr on SOC
	u64	*recv_bufaddr;	// SOC's Recv_buf addr on SOC
	wait_queue_head_t	recvwait;
	unsigned char	*recv_tempbuf;
	int * send_tail;
	int * recv_tail;
	unsigned long buf_start;
	int vnet_id;
};

extern  int evtchn_do_upcall(struct pt_regs *regs);
extern void handle_evt(int evt, struct pt_regs * regs);
extern int request_evtchn(unsigned int evt, 
		evtreturn_t (*handler)(int, void *, struct pt_regs *),unsigned long evtflags,
		const char * evtname, void *evt_id);
extern void free_evt(unsigned long evt,void * evt_id);
extern int hcall_send_vcpu_evt(int evt, int domid, int vcpuid);
extern int hcall_send_evt(int evt,int domid);
extern int get_cg_domain_id(int cgid);
extern int check_vcpu_alive(int domid, int vcpuid);
extern int prepare_shared_info(int domid);

extern void set_vcpu_dead(int domid, int vcpuid);
extern void unlock_vcpu_evtchnlock(int domid, int vcpuid);

#endif   /* ----- #ifndef _SWVMM_EVTCHN_H_  ----- */

