/*
 * =====================================================================================
 *       Filename:  swvmm.h
 *    Description:  
 *        Created:  2008年12月22日 11时38分25秒
 * =====================================================================================
 */


#ifndef  _SWVMM_H_
#define  _SWVMM_H_

#include <linux/types.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/core_sw.h>

typedef uint16_t domid_t;

typedef struct {
	    volatile unsigned int lock;
} sw_spinlock_t;
#define __SW_SPIN_LOCK_UNLOCKED        { 0 }

static inline void sw_spin_unlock(sw_spinlock_t * lock)
{
    __asm__("memb");
	lock->lock = 0;
}

static inline void sw_spin_lock(sw_spinlock_t * lock)
{
        long tmp;
       __asm__ __volatile__(
		"   memb\n"
        "1:     ldi %0,%1\n"
        "       ldw_set %0,0(%0)\n"     //取出锁地址内容，并对锁地址置1
        "       bne     %0,1b\n"        //如果不返回0（为1），则表示锁已被占住  
        "       memb\n"                 //如果返回0，则占用锁，并上锁. 
        : "=&r" (tmp), "=m" (lock->lock)
        : "m"(lock->lock) : "memory");
}


#define sw_spin_lock_init(lock_ptr)					\
	do { (lock_ptr)->lock = 0; } while (0)

static inline unsigned long _sw_spin_lock_irqsave(sw_spinlock_t * lock)
{
        long tmp;
		long flags;
		local_irq_save(flags);
		__asm__ __volatile__(
				"   memb\n"
				"1:     ldi %0,%1\n"
				"       ldw_set %0,0(%0)\n"     //取出锁地址内容，并对锁地址置1
				"       bne     %0,1b\n"        //如果不返回0（为1），则表示锁已被占住  
				"       memb\n"                 //如果返回0，则占用锁，并上锁. 
				: "=&r" (tmp), "=m" (lock->lock)
				: "m"(lock->lock) : "memory");
		return flags;
}
static inline void _sw_spin_unlock_irqrestore(sw_spinlock_t * lock, unsigned long flags)
{
    __asm__("memb");
	lock->lock = 0;
	local_irq_restore(flags);
}

#define sw_spin_lock_irqsave(lock, flags)          \
	do {                        \
		typecheck(unsigned long, flags);    \
		flags = _sw_spin_lock_irqsave(lock);   \
	} while (0)


#define sw_spin_unlock_irqrestore(lock, flags)		\
	do {						\
		typecheck(unsigned long, flags);	\
		_sw_spin_unlock_irqrestore(lock, flags);	\
	} while (0)

#define ___DEFINE_SWVMM_USER_HANDLE(name, type) \
	    typedef type * __guest_handle_ ## name
#define __DEFINE_SWVMM_USER_HANDLE(name, type) \
	    ___DEFINE_SWVMM_USER_HANDLE(name, type)
#define DEFINE_SWVMM_USER_HANDLE(name)   __DEFINE_SWVMM_USER_HANDLE(name, name)
#define __SWVMM_USER_HANDLE(name)        __guest_handle_ ## name
#define SWVMM_USER_HANDLE(name)          __SWVMM_USER_HANDLE(name)


/* Maximum number of virtual CPUs in multi-processor guests. */
#define MAX_VIRT_CPUS 4
#define MAX_DOMAINS 16

#define SWVMM_KERNEL_SHARED_MEM_setup 0
#define SWVMM_KERNEL_SHARED_MEM_map 1
#define SWVMM_KERNEL_SHARED_MEM_map_all 2
#define SWVMM_KERNEL_SHARED_MEM_setup_idx 3
#define SWVMM_KERNEL_SHARED_MEM_del 4
#define INIT_CG_FLAGS 0x1001UL
#define USER_CG_FLAGS 0x1002UL
#define IDX_CG_FLAGS 0x1003UL
struct swvmm_kernel_sharedmem {
	/* IN parameters */
	unsigned long cmd;
	unsigned long flags;
	/* IN/OUT parameters */                               
	unsigned long size;
	unsigned long addr_page_table;                             
	unsigned long index;  
	unsigned long addr_private_buf;  
	unsigned long vma_flags;
	void *domain;
};

struct vcpu_info {
	uint8_t vcpu_alive;  
	uint8_t pcpu_id;
	unsigned long evtchn_pending;
	unsigned long evtchn_pending_stat[64];
	sw_spinlock_t evtchn_lock;
	unsigned long evtchn_mask;
};      

typedef struct vcpu_info vcpu_info_t;


struct vm_call_single_queue {
	struct list_head list;
	sw_spinlock_t lock;
};

#define SWVMM_INIT_DONE_FLAG 0xd
struct shared_info {
	unsigned long soft_core_start;          /** 固定地址为核心虚地址 0xfffffc0000000000UL */
	uint8_t 	is_hotreset;
	uint8_t 	hard_cg_id;
	struct vcpu_info vcpu_info[MAX_VIRT_CPUS];
	struct vm_call_single_queue csq;
	sw_spinlock_t  call_function_lock;
	unsigned long ivmi_data;
	struct list_head call_function_queue;
	unsigned long unused_space[64];
	void* sw3mm_addr;                       /*hgy add: 多cg间共享空间的起始地址*/	
	uint8_t 	is_init_done;           /** SWVMM 相关的初始化完成，可以发送SMVM中断的标志 */
};

typedef struct shared_info shared_info_t;

extern shared_info_t * SWVMM_shared_info;
#define PAL_hard_cgid 0x8F
__CALL_PAL_R0(hard_cgid, unsigned long);

extern int is_hotreset(void);
#define VM_PCI_CG_SHARED    0x80000000  /* the vma is use for CG shared */
struct vma_part_info {
	unsigned long addr;
	unsigned long size;
	unsigned long cgshared;
};
extern unsigned long * domain_shared_info[MAX_DOMAINS];
static inline unsigned long get_vmem_offset(void)
{
	return hwrpb->vmem_offset;
}
static inline unsigned long get_domain_onlinemap(void)
{
	unsigned long bit = 0;
	int i,j = 0;
	for(i=0; i<CG_NUM; i++)
	{
		if(is_cg_alive(i))
		{
			bit |= 1UL << j;
			j ++;
		}
	}
	return bit;
}
static inline shared_info_t * get_domain_sharedinfo(int domid)
{
	return (shared_info_t *) domain_shared_info[domid];
}
static inline int  domain_online(int domid)
{
	unsigned long online = get_domain_onlinemap();
	return test_bit(domid , &online);
}
static inline int get_vcpu_pcpuid(int vcpuid)
{
    struct percpu_struct * cpu;
    cpu = (struct percpu_struct *)
        ((char*)hwrpb
         + hwrpb->processor_offset
         + (vcpuid)*hwrpb->processor_size);
    return (cpu->phy_cpuid);
}

/* SW Vmalloc Related */
extern unsigned long kva2pa_map(unsigned long * pfn_table, unsigned long count);
extern unsigned long kva_map_pfn_range(unsigned long pfn, unsigned long size, pgprot_t prot);

extern void shfree_s(void *obj, int size);
extern void * _swvmalloc(int len);
#define swvfree(x) shfree_s((void *)(x), 0)
#define swvmalloc(len) _swvmalloc(len)


/* SMVM related */
extern int smvm_call_function_single(int domid, void (*func) (void *info), void *info,int wait);
extern int smvm_call_function_mask(unsigned long mask, void (*func)(void *), void *info,int wait);
extern int smvm_call_function(void (*func)(void *), void *info, int wait);

#endif   /* ----- #ifndef _SWVMM_H_  ----- */
