/*
 * =====================================================================================
 *       Filename:  xchg_numa.h
 *    Description:  
 *        Created:  02/08/2013 09:17:59 AM
 * =====================================================================================
 */


#ifndef _SW_CMPXCHG_H
#error Do not include xchg.h directly!
#else

#include <linux/compiler.h>
#include <asm/types.h>		/* for BITS_PER_LONG/SHIFT_PER_LONG */
#include <asm/byteorder.h>
#include <asm/atomic.h>
/*
 * xchg/xchg_local and cmpxchg/cmpxchg_local share the same code
 * except that local version do not have the expensive memory barrier.
 * So this file is included twice from asm/cmpxchg.h.
 */

/*
 * Atomic exchange.
 * Since it can be used to implement critical sections
 * it must clobber "memory" (also for interrupts in UP).
 */

static inline unsigned long
____xchg(_u8, volatile char *m, unsigned long val)
{
	unsigned long ret, tmp, addr64;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);
	__asm__ __volatile__(
	"	andnot	%4,7,%3\n"
	"	ins0b	%1,%4,%1\n"
	"	ldl	%2,0(%3)\n"
	"	ext0b	%2,%4,%0\n"
	"	mask0b	%2,%4,%2\n"
	"	or	%1,%2,%2\n"
	"	stl	%2,0(%3)\n"
		__ASM__MB
	: "=&r" (ret), "=&r" (val), "=&r" (tmp), "=&r" (addr64)
	: "r" ((long)m), "1" (val) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return ret;
}

static inline unsigned long
____xchg(_u16, volatile short *m, unsigned long val)
{
	unsigned long ret, tmp, addr64;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);

	__asm__ __volatile__(
	"	andnot	%4,7,%3\n"
	"	ins1b	%1,%4,%1\n"
	"	ldl	%2,0(%3)\n"
	"	ext1b	%2,%4,%0\n"
	"	mask1b	%2,%4,%2\n"
	"	or	%1,%2,%2\n"
	"	stl	%2,0(%3)\n"
		__ASM__MB
	: "=&r" (ret), "=&r" (val), "=&r" (tmp), "=&r" (addr64)
	: "r" ((long)m), "1" (val) : "memory");
	_atomic_spin_unlock_irqrestore(m, flags);

	return ret;
}

static inline unsigned long
____xchg(_u32, volatile int *m, unsigned long val)
{
	unsigned long dummy;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);

	__asm__ __volatile__(
	"	ldw %0,%4\n"
	"	bis $31,%3,%1\n"
	"	stw %1,%2\n"
		__ASM__MB
	: "=&r" (val), "=&r" (dummy), "=m" (*m)
	: "rI" (val), "m" (*m) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return val;
}

static inline unsigned long
____xchg(_u64, volatile long *m, unsigned long val)
{
	unsigned long dummy;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);
	__asm__ __volatile__(
	"	ldl %0,%4\n"
	"	bis $31,%3,%1\n"
	"	stl %1,%2\n"
		__ASM__MB
	: "=&r" (val), "=&r" (dummy), "=m" (*m)
	: "rI" (val), "m" (*m) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return val;
}

/* This function doesn't exist, so you'll get a linker error
   if something tries to do an invalid xchg().  */
extern void __xchg_called_with_bad_pointer(void);

static __always_inline unsigned long
____xchg(, volatile void *ptr, unsigned long x, int size)
{
	switch (size) {
		case 1:
			return ____xchg(_u8, ptr, x);
		case 2:
			return ____xchg(_u16, ptr, x);
		case 4:
			return ____xchg(_u32, ptr, x);
		case 8:
			return ____xchg(_u64, ptr, x);
	}
	__xchg_called_with_bad_pointer();
	return x;
}

/*
 * Atomic compare and exchange.  Compare OLD with MEM, if identical,
 * store NEW in MEM.  Return the initial value in MEM.  Success is
 * indicated by comparing RETURN with OLD.
 *
 * The memory barrier should be placed in SMP only when we actually
 * make the change. If we don't change anything (so if the returned
 * prev is equal to old) then we aren't acquiring anything new and
 * we don't need any memory barrier as far I can tell.
 */

static inline unsigned long
____cmpxchg(_u8, volatile char *m, unsigned char old, unsigned char new)
{
	unsigned long prev, tmp, cmp, addr64;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);

	__asm__ __volatile__(
	"	andnot	%5,7,%4\n"
	"	ins0b	%1,%5,%1\n"
	"1:	ldl	%2,0(%4)\n"
	"	ext0b	%2,%5,%0\n"
	"	cmpeq	%0,%6,%3\n"
	"	beq	%3,2f\n"
	"	mask0b	%2,%5,%2\n"
	"	or	%1,%2,%2\n"
	"	stl	%2,0(%4)\n"
		__ASM__MB
	"2:\n"
	: "=&r" (prev), "=&r" (new), "=&r" (tmp), "=&r" (cmp), "=&r" (addr64)
	: "r" ((long)m), "Ir" (old), "1" (new) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return prev;
}

static inline unsigned long
____cmpxchg(_u16, volatile short *m, unsigned short old, unsigned short new)
{
	unsigned long prev, tmp, cmp, addr64;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);

	__asm__ __volatile__(
	"	andnot	%5,7,%4\n"
	"	ins1b	%1,%5,%1\n"
	"1:	ldl	%2,0(%4)\n"
	"	ext1b	%2,%5,%0\n"
	"	cmpeq	%0,%6,%3\n"
	"	beq	%3,2f\n"
	"	mask1b	%2,%5,%2\n"
	"	or	%1,%2,%2\n"
	"	stl	%2,0(%4)\n"
		__ASM__MB
	"2:\n"
	: "=&r" (prev), "=&r" (new), "=&r" (tmp), "=&r" (cmp), "=&r" (addr64)
	: "r" ((long)m), "Ir" (old), "1" (new) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return prev;
}

static inline unsigned long
____cmpxchg(_u32, volatile int *m, int old, int new)
{
	unsigned long prev, cmp;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);

	__asm__ __volatile__(
	"	ldw %0,%5\n"
	"	cmpeq %0,%3,%1\n"
	"	beq %1,2f\n"
	"	mov %4,%1\n"
	"	stw %1,%2\n"
		__ASM__MB
	"2:\n"
	: "=&r"(prev), "=&r"(cmp), "=m"(*m)
	: "r"((long) old), "r"(new), "m"(*m) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return prev;
}

static inline unsigned long
____cmpxchg(_u64, volatile long *m, unsigned long old, unsigned long new)
{
	unsigned long prev, cmp;
	unsigned long flags;

	_atomic_spin_lock_irqsave(m, flags);

	__asm__ __volatile__(
	"	ldl %0,%5\n"
	"	cmpeq %0,%3,%1\n"
	"	beq %1,2f\n"
	"	mov %4,%1\n"
	"	stl %1,%2\n"
		__ASM__MB
	"2:\n"
	: "=&r"(prev), "=&r"(cmp), "=m"(*m)
	: "r"((long) old), "r"(new), "m"(*m) : "memory");

	_atomic_spin_unlock_irqrestore(m, flags);
	return prev;
}

/* This function doesn't exist, so you'll get a linker error
   if something tries to do an invalid cmpxchg().  */
extern void __cmpxchg_called_with_bad_pointer(void);

static __always_inline unsigned long
____cmpxchg(, volatile void *ptr, unsigned long old, unsigned long new,
	      int size)
{
	switch (size) {
		case 1:
			return ____cmpxchg(_u8, ptr, old, new);
		case 2:
			return ____cmpxchg(_u16, ptr, old, new);
		case 4:
			return ____cmpxchg(_u32, ptr, old, new);
		case 8:
			return ____cmpxchg(_u64, ptr, old, new);
	}
	__cmpxchg_called_with_bad_pointer();
	return old;
}

#endif
